import lightning


def test_model_trainer_fit(multimodal_model, sample_train_val_datamodule):
    trainer = lightning.pytorch.trainer.trainer.Trainer(fast_dev_run=True, accelerator="cpu")
    trainer.fit(model=multimodal_model, datamodule=sample_train_val_datamodule)



"""Model which uses mutliple prediction heads"""
from typing import Optional

import torch
from torch import nn

from pvnet.models.base_model import BaseModel


class Ensemble(BaseModel):
    """Ensemble of PVNet models"""

    def __init__(
        self,
        model_list: list[BaseModel],
        weights: Optional[list[float]] = None,
    ):
        """Ensemble of PVNet models

        Args:
            model_list: A list of PVNet models to ensemble
            weights: A list of weighting to apply to each model. If None, the models are weighted
                equally.
        """

        # Surface check all the models are compatible
        output_quantiles = []
        history_minutes = []
        forecast_minutes = []
        target_key = []
        interval_minutes = []

        # Get some model properties from each model
        for model in model_list:
            output_quantiles.append(model.output_quantiles)
            history_minutes.append(model.history_minutes)
            forecast_minutes.append(model.forecast_minutes)
            target_key.append(model._target_key_name)
            interval_minutes.append(model.interval_minutes)

        # Check these properties are all the same
        for param_list in [
            output_quantiles,
            history_minutes,
            forecast_minutes,
            target_key,
            interval_minutes,
        ]:
            assert all([p == param_list[0] for p in param_list]), param_list

        super().__init__(
            history_minutes=history_minutes[0],
            forecast_minutes=forecast_minutes[0],
            optimizer=None,
            output_quantiles=output_quantiles[0],
            target_key=target_key[0],
            interval_minutes=interval_minutes[0],
        )

        self.model_list = nn.ModuleList(model_list)

        if weights is None:
            weights = torch.ones(len(model_list)) / len(model_list)
        else:
            assert len(weights) == len(model_list)
            weights = torch.Tensor(weights) / sum(weights)
        self.weights = nn.Parameter(weights, requires_grad=False)

    def forward(self, batch):
        """Run the model forward"""
        y_hat = 0
        for weight, model in zip(self.weights, self.model_list):
            y_hat = model(batch) * weight + y_hat
        return y_hat



"""Linear networks used for the fusion model"""
from torch import nn, rand

from pvnet.models.multimodal.linear_networks.basic_blocks import (
    AbstractLinearNetwork,
    ResidualLinearBlock,
    ResidualLinearBlock2,
)


class DefaultFCNet(AbstractLinearNetwork):
    """Similar to the original FCNet module used in PVNet, with a few minor tweaks.

    This is a 2-layer fully connected block, with internal ELU activations and output ReLU.
    """

    def __init__(
        self,
        in_features: int,
        out_features: int,
        fc_hidden_features: int = 128,
    ):
        """Similar to the original FCNet module used in PVNet, with a few minor tweaks.

        Args:
            in_features: Number of input features.
            out_features: Number of output features.
            fc_hidden_features: Number of features in middle hidden layer.
        """
        super().__init__(in_features, out_features)

        self.model = nn.Sequential(
            nn.Linear(in_features=in_features, out_features=fc_hidden_features),
            nn.ELU(),
            nn.Linear(in_features=fc_hidden_features, out_features=out_features),
            nn.ReLU(),
        )

    def forward(self, x):
        """Run model forward"""
        x = self.cat_modes(x)
        return self.model(x)


class ResFCNet(AbstractLinearNetwork):
    """Fully-connected deep network based on ResNet architecture.

    Internally, this network uses ELU activations throughout the residual blocks.
    With n_res_blocks=0 this becomes equivalent to `DefaultFCNet`.
    """

    def __init__(
        self,
        in_features: int,
        out_features: int,
        fc_hidden_features: int = 128,
        n_res_blocks: int = 4,
        res_block_layers: int = 2,
        dropout_frac: float = 0.2,
    ):
        """Fully-connected deep network based on ResNet architecture.

        Args:
            in_features: Number of input features.
            out_features: Number of output features.
            fc_hidden_features: Number of features in middle hidden layers.
            n_res_blocks: Number of residual blocks to use.
            res_block_layers: Number of fully-connected layers used in each residual block.
            dropout_frac: Probability of an element to be zeroed in the residual pathways.
        """
        super().__init__(in_features, out_features)

        model = [
            nn.Linear(in_features=in_features, out_features=fc_hidden_features),
        ]

        for i in range(n_res_blocks):
            model += [
                ResidualLinearBlock(
                    in_features=fc_hidden_features,
                    n_layers=res_block_layers,
                    dropout_frac=dropout_frac,
                )
            ]

        model += [
            nn.ELU(),
            nn.Linear(in_features=fc_hidden_features, out_features=out_features),
            nn.LeakyReLU(negative_slope=0.01),
        ]
        self.model = nn.Sequential(*model)

    def forward(self, x):
        """Run model forward"""
        x = self.cat_modes(x)
        return self.model(x)


class ResFCNet2(AbstractLinearNetwork):
    """Fully connected deep network based on ResNet architecture.

    This architecture is similar to
    `ResFCNet`, except that it uses LeakyReLU activations internally, and batchnorm in the residual
    branches. The residual blocks are implemented based on the best performing block in [1].

    Sources:
        [1] https://arxiv.org/pdf/1603.05027.pdf
    """

    def __init__(
        self,
        in_features: int,
        out_features: int,
        fc_hidden_features: int = 128,
        n_res_blocks: int = 4,
        res_block_layers: int = 2,
        dropout_frac=0.0,
    ):
        """Fully connected deep network based on ResNet architecture.

        Args:
            in_features: Number of input features.
            out_features: Number of output features.
            fc_hidden_features: Number of features in middle hidden layers.
            n_res_blocks: Number of residual blocks to use.
            res_block_layers: Number of fully-connected layers used in each residual block.
            dropout_frac: Probability of an element to be zeroed in the residual pathways.
        """
        super().__init__(in_features, out_features)

        model = [
            nn.Linear(in_features=in_features, out_features=fc_hidden_features),
        ]

        for i in range(n_res_blocks):
            model += [
                ResidualLinearBlock2(
                    in_features=fc_hidden_features,
                    n_layers=res_block_layers,
                    dropout_frac=dropout_frac,
                )
            ]

        model += [
            nn.LeakyReLU(),
            nn.Linear(in_features=fc_hidden_features, out_features=out_features),
            nn.LeakyReLU(negative_slope=0.01),
        ]

        self.model = nn.Sequential(*model)

    def forward(self, x):
        """Run model forward"""
        x = self.cat_modes(x)
        return self.model(x)


class SNN(AbstractLinearNetwork):
    """Self normalising neural network implementation borrowed from [1] and proposed in [2].

    Sources:
        [1] https://github.com/tonyduan/snn/blob/master/snn/models.py
        [2] https://arxiv.org/pdf/1706.02515v5.pdf

    Args:
        in_features: Number of input features.
        out_features: Number of output features.
        fc_hidden_features: Number of features in middle hidden layers.
        n_layers: Number of fully-connected layers used in the network.
        dropout_frac: Probability of an element to be zeroed.

    """

    def __init__(
        self,
        in_features: int,
        out_features: int,
        fc_hidden_features: int = 128,
        n_layers: int = 10,
        dropout_frac: float = 0.0,
    ):
        """Self normalising neural network implementation borrowed from [1] and proposed in [2].

        Sources:
            [1] https://github.com/tonyduan/snn/blob/master/snn/models.py
            [2] https://arxiv.org/pdf/1706.02515v5.pdf

        Args:
            in_features: Number of input features.
            out_features: Number of output features.
            fc_hidden_features: Number of features in middle hidden layers.
            n_layers: Number of fully-connected layers used in the network.
            dropout_frac: Probability of an element to be zeroed.

        """
        super().__init__(in_features, out_features)

        layers = [
            nn.Linear(in_features, fc_hidden_features, bias=False),
            nn.SELU(),
            nn.AlphaDropout(p=dropout_frac),
        ]
        for i in range(1, n_layers - 1):
            layers += [
                nn.Linear(fc_hidden_features, fc_hidden_features, bias=False),
                nn.SELU(),
                nn.AlphaDropout(p=dropout_frac),
            ]
        layers += [
            nn.Linear(fc_hidden_features, out_features, bias=True),
            nn.LeakyReLU(negative_slope=0.01),
        ]

        self.network = nn.Sequential(*layers)
        self._reset_parameters()

    def forward(self, x):
        """Run model forward"""
        x = self.cat_modes(x)
        return self.network(x)

    def _reset_parameters(self):
        for layer in self.network:
            if isinstance(layer, nn.Linear):
                nn.init.normal_(layer.weight, std=layer.out_features**-0.5)
                if layer.bias is not None:
                    fan_in, _ = nn.init._calculate_fan_in_and_fan_out(layer.weight)
                    bound = fan_in**-0.5
                    nn.init.uniform_(layer.bias, -bound, bound)


class TabNet(AbstractLinearNetwork):
    """An implmentation of TabNet [1].

    The implementation comes rom `pytorch_tabnet` and this must be installed for use.


    Sources:
        [1] https://arxiv.org/abs/1908.07442
    """

    def __init__(
        self,
        in_features: int,
        out_features: int,
        n_d=8,
        n_a=8,
        n_steps=3,
        gamma=1.3,
        cat_idxs=[],
        cat_dims=[],
        cat_emb_dim=1,
        n_independent=2,
        n_shared=2,
        epsilon=1e-15,
        virtual_batch_size=128,
        momentum=0.02,
        mask_type="sparsemax",
    ):
        """An implmentation of TabNet [1].

        Sources:
            [1] https://arxiv.org/abs/1908.07442

        Args:
            in_features: int
                Number of input features.
            out_features: int
                Number of output features.
            n_d : int
                Dimension of the prediction  layer (usually between 4 and 64)
            n_a : int
                Dimension of the attention  layer (usually between 4 and 64)
            n_steps : int
                Number of successive steps in the network (usually between 3 and 10)
            gamma : float
                Float above 1, scaling factor for attention updates (usually between 1.0 to 2.0)
            cat_idxs : list of int
                Index of each categorical column in the dataset
            cat_dims : list of int
                Number of categories in each categorical column
            cat_emb_dim : int or list of int
                Size of the embedding of categorical features
                if int, all categorical features will have same embedding size
                if list of int, every corresponding feature will have specific size
            n_independent : int
                Number of independent GLU layer in each GLU block (default 2)
            n_shared : int
                Number of independent GLU layer in each GLU block (default 2)
            epsilon : float
                Avoid log(0), this should be kept very low
            virtual_batch_size : int
                Batch size for Ghost Batch Normalization
            momentum : float
                Float value between 0 and 1 which will be used for momentum in all batch norm
            mask_type : str
                Either "sparsemax" or "entmax" : this is the masking function to use
        """
        from pytorch_tabnet.tab_network import TabNet as _TabNetModel

        super().__init__(in_features, out_features)

        self._tabnet = _TabNetModel(
            input_dim=in_features,
            output_dim=out_features,
            n_d=n_d,
            n_a=n_a,
            n_steps=n_steps,
            gamma=gamma,
            cat_idxs=cat_idxs,
            cat_dims=cat_dims,
            cat_emb_dim=cat_emb_dim,
            n_independent=n_independent,
            n_shared=n_shared,
            epsilon=epsilon,
            virtual_batch_size=virtual_batch_size,
            momentum=momentum,
            mask_type=mask_type,
            group_attention_matrix=rand(4, in_features),
        )

        self.activation = nn.LeakyReLU(negative_slope=0.01)

    def forward(self, x):
        """Run model forward"""
        # TODO: USE THIS LOSS COMPONENT
        # loss = self.compute_loss(output, y)
        # Add the overall sparsity loss
        # loss = loss - self.lambda_sparse * M_loss
        x = self.cat_modes(x)
        out1, M_loss = self._tabnet(x)
        return self.activation(out1)



"""Basic blocks for the lienar networks"""
from abc import ABCMeta, abstractmethod
from collections import OrderedDict

import torch
from torch import nn


class AbstractLinearNetwork(nn.Module, metaclass=ABCMeta):
    """Abstract class for a network to combine the features from all the inputs."""

    def __init__(
        self,
        in_features: int,
        out_features: int,
    ):
        """Abstract class for a network to combine the features from all the inputs.

        Args:
            in_features: Number of input features.
            out_features: Number of output features.
        """
        super().__init__()

    def cat_modes(self, x):
        """Concatenate modes of input data into 1D feature vector"""
        if isinstance(x, OrderedDict):
            return torch.cat([value for key, value in x.items()], dim=1)
        elif isinstance(x, torch.Tensor):
            return x
        else:
            raise ValueError(f"Input of unexpected type {type(x)}")

    @abstractmethod
    def forward(self):
        """Run model forward"""
        pass


class ResidualLinearBlock(nn.Module):
    """A 1D fully-connected residual block using ELU activations and including optional dropout."""

    def __init__(
        self,
        in_features: int,
        n_layers: int = 2,
        dropout_frac: float = 0.0,
    ):
        """A 1D fully-connected residual block using ELU activations and including optional dropout.

        Args:
            in_features: Number of input features.
            n_layers: Number of layers in residual pathway.
            dropout_frac: Probability of an element to be zeroed.
        """
        super().__init__()

        layers = []
        for i in range(n_layers):
            layers += [
                nn.ELU(),
                nn.Linear(
                    in_features=in_features,
                    out_features=in_features,
                ),
                nn.Dropout(p=dropout_frac),
            ]
        self.model = nn.Sequential(*layers)

    def forward(self, x):
        """Run model forward"""
        return self.model(x) + x


class ResidualLinearBlock2(nn.Module):
    """Residual block of 'full pre-activation' similar to the block in figure 4(e) of [1].

    This was the best performing residual block tested in the study. This implementation differs
    from that block just by using LeakyReLU activation to avoid dead neuron, and by including
    optional dropout in the residual branch. This is also a 1D fully connected layer residual block
    rather than a 2D convolutional block.

    Sources:
        [1] https://arxiv.org/pdf/1603.05027.pdf
    """

    def __init__(
        self,
        in_features: int,
        n_layers: int = 2,
        dropout_frac: float = 0.0,
    ):
        """Residual block of 'full pre-activation' similar to the block in figure 4(e) of [1].

        Sources:
            [1] https://arxiv.org/pdf/1603.05027.pdf

        Args:
            in_features: Number of input features.
            n_layers: Number of layers in residual pathway.
            dropout_frac: Probability of an element to be zeroed.
        """
        super().__init__()

        layers = []
        for i in range(n_layers):
            layers += [
                nn.BatchNorm1d(in_features),
                nn.Dropout(p=dropout_frac),
                nn.LeakyReLU(),
                nn.Linear(
                    in_features=in_features,
                    out_features=in_features,
                ),
            ]

        self.model = nn.Sequential(*layers)

    def forward(self, x):
        """Run model forward"""
        return self.model(x) + x



"""The default composite model architecture for PVNet"""

import glob
from collections import OrderedDict
from typing import Optional

import hydra
import torch
import torch.nn.functional as F
from ocf_datapipes.batch import BatchKey, NWPBatchKey
from pyaml_env import parse_config
from torch import nn

import pvnet
from pvnet.models.multimodal.linear_networks.basic_blocks import AbstractLinearNetwork
from pvnet.models.multimodal.multimodal_base import MultimodalBaseModel
from pvnet.optimizers import AbstractOptimizer


class Model(MultimodalBaseModel):
    """Neural network which combines information from different sources

    The network is trained via unimodal teachers [1].

    Architecture is roughly as follows:

    - Satellite data, if included, is put through an encoder which transforms it from 4D, with time,
        channel, height, and width dimensions to become a 1D feature vector.
    - NWP, if included, is put through a similar encoder.
    - PV site-level data, if included, is put through an encoder which transforms it from 2D, with
        time and system-ID dimensions, to become a 1D feature vector.
    - The satellite features*, NWP features*, PV site-level features*, GSP ID embedding*, and sun
        paramters* are concatenated into a 1D feature vector and passed through another neural
        network to combine them and produce a forecast.

    * if included
    [1] https://arxiv.org/pdf/2305.01233.pdf
    """

    name = "unimodal_teacher"

    def __init__(
        self,
        output_network: AbstractLinearNetwork,
        output_quantiles: Optional[list[float]] = None,
        include_gsp_yield_history: bool = True,
        include_sun: bool = True,
        embedding_dim: Optional[int] = 16,
        forecast_minutes: int = 30,
        history_minutes: int = 60,
        optimizer: AbstractOptimizer = pvnet.optimizers.Adam(),
        mode_teacher_dict: dict = {},
        val_best: bool = True,
        cold_start: bool = True,
        enc_loss_frac: float = 0.3,
        adapt_batches: Optional[bool] = False,
    ):
        """Neural network which combines information from different sources.

        The network is trained via unimodal teachers [1].

        [1] https://arxiv.org/pdf/2305.01233.pdf

        Notes:
            In the args, where it says a module `m` is partially instantiated, it means that a
            normal pytorch module will be returned by running `mod = m(**kwargs)`. In this library,
            this partial instantiation is generally achieved using partial instantiation via hydra.
            However, the arg is still valid as long as `m(**kwargs)` returns a valid pytorch module
            - for example if `m` is a regular function.

        Args:
            output_network: A partially instatiated pytorch Module class used to combine the 1D
                features to produce the forecast.
            output_quantiles: A list of float (0.0, 1.0) quantiles to predict values for. If set to
                None the output is a single value.
            include_gsp_yield_history: Include GSP yield data.
            include_sun: Include sun azimuth and altitude data.
            embedding_dim: Number of embedding dimensions to use for GSP ID. Not included if set to
                `None`.
            forecast_minutes: The amount of minutes that should be forecasted.
            history_minutes: The default amount of historical minutes that are used.
            optimizer: Optimizer factory function used for network.
            mode_teacher_dict: A dictionary of paths to different model checkpoint directories,
                which will be used as the unimodal teachers.
            val_best: Whether to load the model which performed best on the validation set. Else the
                last checkpoint is loaded.
            cold_start: Whether to train the uni-modal encoders from scratch. Else start them with
                weights from the uni-modal teachers.
            enc_loss_frac: Fraction of total loss attributed to the teacher encoders.
            adapt_batches: If set to true, we attempt to slice the batches to the expected shape for
                the model to use. This allows us to overprepare batches and slice from them for the
                data we need for a model run.
        """

        self.include_gsp_yield_history = include_gsp_yield_history
        self.include_sun = include_sun
        self.embedding_dim = embedding_dim
        self.enc_loss_frac = enc_loss_frac
        self.include_sat = False
        self.include_nwp = False
        self.include_pv = False
        self.adapt_batches = adapt_batches

        # This is set but modified later based on the teachers
        self.add_image_embedding_channel = False

        super().__init__(
            history_minutes=history_minutes,
            forecast_minutes=forecast_minutes,
            optimizer=optimizer,
            output_quantiles=output_quantiles,
            target_key="gsp",
        )

        # Number of features expected by the output_network
        # Add to this as network pices are constructed
        fusion_input_features = 0

        self.teacher_models = torch.nn.ModuleDict()
        self.mode_teacher_dict = mode_teacher_dict

        for mode, path in mode_teacher_dict.items():
            # load teacher model and freeze its weights
            self.teacher_models[mode] = self.get_unimodal_encoder(path, True, val_best=val_best)

            for param in self.teacher_models[mode].parameters():
                param.requires_grad = False

            # Recreate model as student
            mode_student_model = self.get_unimodal_encoder(
                path, load_weights=(not cold_start), val_best=val_best
            )

            if mode == "sat":
                self.include_sat = True
                self.sat_sequence_len = mode_student_model.sat_sequence_len
                self.sat_encoder = mode_student_model.sat_encoder

                if mode_student_model.add_image_embedding_channel:
                    self.sat_embed = mode_student_model.sat_embed
                    self.add_image_embedding_channel = True

                fusion_input_features += self.sat_encoder.out_features

            elif mode == "pv":
                self.include_pv = True
                self.pv_encoder = mode_student_model.pv_encoder
                fusion_input_features += self.pv_encoder.out_features

            elif mode.startswith("nwp"):
                nwp_source = mode.removeprefix("nwp/")

                if not self.include_nwp:
                    self.include_nwp = True
                    self.nwp_encoders_dict = torch.nn.ModuleDict()

                    if mode_student_model.add_image_embedding_channel:
                        self.add_image_embedding_channel = True
                        self.nwp_embed_dict = torch.nn.ModuleDict()

                self.nwp_encoders_dict[nwp_source] = mode_student_model.nwp_encoders_dict[
                    nwp_source
                ]

                if self.add_image_embedding_channel:
                    self.nwp_embed_dict[nwp_source] = mode_student_model.nwp_embed_dict[nwp_source]

                fusion_input_features += self.nwp_encoders_dict[nwp_source].out_features

        if self.embedding_dim:
            self.embed = nn.Embedding(num_embeddings=318, embedding_dim=embedding_dim)
            fusion_input_features += embedding_dim

        if self.include_sun:
            self.sun_fc1 = nn.Linear(
                in_features=2 * (self.forecast_len + self.history_len + 1),
                out_features=16,
            )
            fusion_input_features += 16

        if include_gsp_yield_history:
            fusion_input_features += self.history_len

        self.output_network = output_network(
            in_features=fusion_input_features,
            out_features=self.num_output_features,
        )

        self.save_hyperparameters()

    def get_unimodal_encoder(self, path, load_weights, val_best):
        """Load a model to function as a unimodal teacher"""

        model_config = parse_config(f"{path}/model_config.yaml")

        # Load the teacher model
        encoder = hydra.utils.instantiate(model_config)

        if load_weights:
            if val_best:
                # Only one epoch (best) saved per model
                files = glob.glob(f"{path}/epoch*.ckpt")
                assert len(files) == 1
                checkpoint = torch.load(files[0], map_location="cpu")
            else:
                checkpoint = torch.load(f"{path}/last.ckpt", map_location="cpu")

            encoder.load_state_dict(state_dict=checkpoint["state_dict"])
        return encoder

    def teacher_forward(self, x):
        """Run the teacher models and return their encodings"""
        modes = OrderedDict()
        for mode, teacher_model in self.teacher_models.items():
            # ******************* Satellite imagery *************************
            if mode == "sat":
                # Shape: batch_size, seq_length, channel, height, width
                sat_data = x[BatchKey.satellite_actual][:, : teacher_model.sat_sequence_len]
                sat_data = torch.swapaxes(sat_data, 1, 2).float()  # switch time and channels

                if self.add_image_embedding_channel:
                    id = x[BatchKey.gsp_id][:, 0].int()
                    sat_data = teacher_model.sat_embed(sat_data, id)

                modes[mode] = teacher_model.sat_encoder(sat_data)

            # *********************** NWP Data ************************************
            if mode.startswith("nwp"):
                nwp_source = mode.removeprefix("nwp/")

                # shape: batch_size, seq_len, n_chans, height, width
                nwp_data = x[BatchKey.nwp][nwp_source][NWPBatchKey.nwp].float()
                nwp_data = torch.swapaxes(nwp_data, 1, 2)  # switch time and channels
                nwp_data = torch.clip(nwp_data, min=-50, max=50)
                if teacher_model.add_image_embedding_channel:
                    id = x[BatchKey.gsp_id][:, 0].int()
                    nwp_data = teacher_model.nwp_embed_dict[nwp_source](nwp_data, id)

                nwp_out = teacher_model.nwp_encoders_dict[nwp_source](nwp_data)
                modes[mode] = nwp_out

            # *********************** PV Data *************************************
            # Add site-level PV yield
            if mode == "pv":
                modes[mode] = teacher_model.pv_encoder(x)

        return modes

    def forward(self, x, return_modes=False):
        """Run model forward"""

        if self.adapt_batches:
            x = self._adapt_batch(x)

        modes = OrderedDict()
        # ******************* Satellite imagery *************************
        if self.include_sat:
            # Shape: batch_size, seq_length, channel, height, width
            sat_data = x[BatchKey.satellite_actual][:, : self.sat_sequence_len]
            sat_data = torch.swapaxes(sat_data, 1, 2).float()  # switch time and channels

            if self.add_image_embedding_channel:
                id = x[BatchKey.gsp_id][:, 0].int()
                sat_data = self.sat_embed(sat_data, id)
            modes["sat"] = self.sat_encoder(sat_data)

        # *********************** NWP Data ************************************
        if self.include_nwp:
            # Loop through potentially many NMPs
            for nwp_source in self.nwp_encoders_dict:
                # shape: batch_size, seq_len, n_chans, height, width
                nwp_data = x[BatchKey.nwp][nwp_source][NWPBatchKey.nwp].float()
                nwp_data = torch.swapaxes(nwp_data, 1, 2)  # switch time and channels
                # Some NWP variables can overflow into NaNs when normalised if they have extreme
                # tails
                nwp_data = torch.clip(nwp_data, min=-50, max=50)

                if self.add_image_embedding_channel:
                    id = x[BatchKey.gsp_id][:, 0].int()
                    nwp_data = self.nwp_embed_dict[nwp_source](nwp_data, id)

                nwp_out = self.nwp_encoders_dict[nwp_source](nwp_data)
                modes[f"nwp/{nwp_source}"] = nwp_out

        # *********************** PV Data *************************************
        # Add site-level PV yield
        if self.include_pv:
            if self._target_key_name != "pv":
                modes["pv"] = self.pv_encoder(x)
            else:
                # Target is PV, so only take the history
                pv_history = x[BatchKey.pv][:, : self.history_len].float()
                modes["pv"] = self.pv_encoder(pv_history)

        # *********************** GSP Data ************************************
        # add gsp yield history
        if self.include_gsp_yield_history:
            gsp_history = x[BatchKey.gsp][:, : self.history_len].float()
            gsp_history = gsp_history.reshape(gsp_history.shape[0], -1)
            modes["gsp"] = gsp_history

        # ********************** Embedding of GSP ID ********************
        if self.embedding_dim:
            id = x[BatchKey.gsp_id][:, 0].int()
            id_embedding = self.embed(id)
            modes["id"] = id_embedding

        if self.include_sun:
            sun = torch.cat(
                (
                    x[BatchKey.gsp_solar_azimuth],
                    x[BatchKey.gsp_solar_elevation],
                ),
                dim=1,
            ).float()
            sun = self.sun_fc1(sun)
            modes["sun"] = sun

        out = self.output_network(modes)

        if self.use_quantile_regression:
            # Shape: batch_size, seq_length * num_quantiles
            out = out.reshape(out.shape[0], self.forecast_len, len(self.output_quantiles))

        if return_modes:
            return out, modes
        else:
            return out

    def _calculate_teacher_loss(self, modes, teacher_modes):
        enc_losses = {}
        for m, enc in teacher_modes.items():
            enc_losses[f"enc_loss/{m}"] = F.l1_loss(enc, modes[m])
        enc_losses["enc_loss/total"] = sum([v for k, v in enc_losses.items()])
        return enc_losses

    def training_step(self, batch, batch_idx):
        """Run training step"""
        y_hat, modes = self.forward(batch, return_modes=True)
        y = batch[self._target_key][:, -self.forecast_len :, 0]

        losses = self._calculate_common_losses(y, y_hat)

        teacher_modes = self.teacher_forward(batch)
        teacher_loss = self._calculate_teacher_loss(modes, teacher_modes)
        losses.update(teacher_loss)

        if self.use_quantile_regression:
            opt_target = losses["quantile_loss"]
        else:
            opt_target = losses["MAE"]

        t_loss = teacher_loss["enc_loss/total"]

        # The scales of the two losses
        l_s = opt_target.detach()
        tl_s = max(t_loss.detach(), 1e-9)

        # opt_target = t_loss/tl_s * l_s * self.enc_loss_frac + opt_target * (1-self.enc_loss_frac)
        losses["opt_loss"] = t_loss / tl_s * l_s * self.enc_loss_frac + opt_target * (
            1 - self.enc_loss_frac
        )

        losses = {f"{k}/train": v for k, v in losses.items()}
        self._training_accumulate_log(batch, batch_idx, losses, y_hat)

        return losses["opt_loss/train"]

    def convert_to_multimodal_model(self, config):
        """Convert the model into a multimodal model class whilst preserving weights"""
        config = config.copy()

        if "cold_start" in config:
            del config["cold_start"]

        config["_target_"] = "pvnet.models.multimodal.multimodal.Model"

        sources = []
        for mode, path in config["mode_teacher_dict"].items():
            model_config = parse_config(f"{path}/model_config.yaml")

            if mode.startswith("nwp"):
                nwp_source = mode.removeprefix("nwp/")
                if "nwp_encoders_dict" in config:
                    for key in ["nwp_encoders_dict", "nwp_history_minutes", "nwp_forecast_minutes"]:
                        config[key][nwp_source] = model_config[key][nwp_source]
                    sources.append("nwp")
                else:
                    for key in ["nwp_encoders_dict", "nwp_history_minutes", "nwp_forecast_minutes"]:
                        config[key] = {nwp_source: model_config[key][nwp_source]}
                config["add_image_embedding_channel"] = model_config["add_image_embedding_channel"]

            elif mode == "sat":
                for key in [
                    "sat_encoder",
                    "add_image_embedding_channel",
                    "min_sat_delay_minutes",
                    "sat_history_minutes",
                ]:
                    config[key] = model_config[key]
                sources.append("sat")

            elif mode == "pv":
                for key in ["pv_encoder", "pv_history_minutes"]:
                    config[key] = model_config[key]
                sources.append("pv")

        del config["mode_teacher_dict"]

        # Load the teacher model
        multimodal_model = hydra.utils.instantiate(config)

        if "sat" in sources:
            multimodal_model.sat_encoder.load_state_dict(self.sat_encoder.state_dict())
        if "nwp" in sources:
            multimodal_model.nwp_encoders_dict.load_state_dict(self.nwp_encoders_dict.state_dict())
        if "pv" in sources:
            multimodal_model.pv_encoder.load_state_dict(self.pv_encoder.state_dict())

        multimodal_model.output_network.load_state_dict(self.output_network.state_dict())

        if self.embedding_dim:
            multimodal_model.embed.load_state_dict(self.embed.state_dict())

        if self.include_sun:
            multimodal_model.sun_fc1.load_state_dict(self.sun_fc1.state_dict())

        return multimodal_model, config



from ocf_datapipes.utils.location import Location
from pvnet.utils import SiteLocationLookup
import xarray as xr
import pytest


@pytest.mark.parametrize(
    "lookup_site_id, expected_x, expected_y, expected_id",
    [
        (1, -1.99106, 48.709865, 1),
        (0, 0.30693, 51.509865, 0),
        (2, -1.56106, 56.203865, 2),
    ],
)
def test_site_location_lookup(lookup_site_id, expected_x, expected_y, expected_id):
    # setup
    site_ids = [0, 1, 2]
    longs = [0.30693, -1.99106, -1.56106]
    lats = [51.509865, 48.709865, 56.203865]
    da_long = xr.DataArray(
        data=longs,
        dims="pv_system_id",
        coords=dict(site_id=(["pv_system_id"], site_ids), long=(["pv_system_id"], longs)),
    )
    da_lat = xr.DataArray(
        data=lats,
        dims="pv_system_id",
        coords=dict(site_id=(["pv_system_id"], site_ids), long=(["pv_system_id"], lats)),
    )
    # Actual testing part
    site_lookup = SiteLocationLookup(long=da_long, lat=da_lat)

    # retrieve location of site 1
    site_location: Location = site_lookup(lookup_site_id)

    assert site_location.x == expected_x
    assert site_location.y == expected_y
    assert site_location.id == expected_id



"""Encoder modules for the satellite/NWP data based on 3D concolutions.
"""
from typing import List, Union

import torch
from torch import nn
from torchvision.transforms import CenterCrop

from pvnet.models.multimodal.encoders.basic_blocks import (
    AbstractNWPSatelliteEncoder,
    ResidualConv3dBlock,
    ResidualConv3dBlock2,
)


class DefaultPVNet(AbstractNWPSatelliteEncoder):
    """This is the original encoding module used in PVNet, with a few minor tweaks."""

    def __init__(
        self,
        sequence_length: int,
        image_size_pixels: int,
        in_channels: int,
        out_features: int,
        number_of_conv3d_layers: int = 4,
        conv3d_channels: int = 32,
        fc_features: int = 128,
        spatial_kernel_size: int = 3,
        temporal_kernel_size: int = 3,
        padding: Union[int, List[int]] = (1, 0, 0),
    ):
        """This is the original encoding module used in PVNet, with a few minor tweaks.

        Args:
            sequence_length: The time sequence length of the data.
            image_size_pixels: The spatial size of the image. Assumed square.
            in_channels: Number of input channels.
            out_features: Number of output features.
            number_of_conv3d_layers: Number of convolution 3d layers that are used.
            conv3d_channels: Number of channels used in each conv3d layer.
            fc_features: number of output nodes out of the hidden fully connected layer.
            spatial_kernel_size: The spatial size of the kernel used in the conv3d layers.
            temporal_kernel_size: The temporal size of the kernel used in the conv3d layers.
            padding: The padding used in the conv3d layers. If an int, the same padding
                is used in all dimensions
        """
        super().__init__(sequence_length, image_size_pixels, in_channels, out_features)
        if isinstance(padding, int):
            padding = (padding, padding, padding)
        # Check that the output shape of the convolutional layers will be at least 1x1
        cnn_spatial_output_size = (
            image_size_pixels
            - ((spatial_kernel_size - 2 * padding[1]) - 1) * number_of_conv3d_layers
        )
        cnn_sequence_length = (
            sequence_length
            - ((temporal_kernel_size - 2 * padding[0]) - 1) * number_of_conv3d_layers
        )
        if not (cnn_spatial_output_size >= 1):
            raise ValueError(
                f"cannot use this many conv3d layers ({number_of_conv3d_layers}) with this input "
                f"spatial size ({image_size_pixels})"
            )

        conv_layers = []

        conv_layers += [
            nn.Conv3d(
                in_channels=in_channels,
                out_channels=conv3d_channels,
                kernel_size=(temporal_kernel_size, spatial_kernel_size, spatial_kernel_size),
                padding=padding,
            ),
            nn.ELU(),
        ]
        for i in range(0, number_of_conv3d_layers - 1):
            conv_layers += [
                nn.Conv3d(
                    in_channels=conv3d_channels,
                    out_channels=conv3d_channels,
                    kernel_size=(temporal_kernel_size, spatial_kernel_size, spatial_kernel_size),
                    padding=padding,
                ),
                nn.ELU(),
            ]

        self.conv_layers = nn.Sequential(*conv_layers)

        # Calculate the size of the output of the 3D convolutional layers
        cnn_output_size = conv3d_channels * cnn_spatial_output_size**2 * cnn_sequence_length

        self.final_block = nn.Sequential(
            nn.Linear(in_features=cnn_output_size, out_features=fc_features),
            nn.ELU(),
            nn.Linear(in_features=fc_features, out_features=out_features),
            nn.ELU(),
        )

    def forward(self, x):
        """Run model forward"""
        out = self.conv_layers(x)
        out = out.reshape(x.shape[0], -1)

        # Fully connected layers
        out = self.final_block(out)

        return out


class DefaultPVNet2(AbstractNWPSatelliteEncoder):
    """The original encoding module used in PVNet, with a few minor tweaks, and batchnorm."""

    def __init__(
        self,
        sequence_length: int,
        image_size_pixels: int,
        in_channels: int,
        out_features: int,
        number_of_conv3d_layers: int = 4,
        conv3d_channels: int = 32,
        fc_features: int = 128,
        batch_norm=True,
        fc_dropout=0.2,
    ):
        """The original encoding module used in PVNet, with a few minor tweaks, and batchnorm.

        Args:
            sequence_length: The time sequence length of the data.
            image_size_pixels: The spatial size of the image. Assumed square.
            in_channels: Number of input channels.
            out_features: Number of output features.
            number_of_conv3d_layers: Number of convolution 3d layers that are used.
            conv3d_channels: Number of channels used in each conv3d layer.
            fc_features: number of output nodes out of the hidden fully connected layer.
            batch_norm: Whether to include 3D batch normalisation.
            fc_dropout: Probability of an element to be zeroed before the last two fully connected
                layers.
        """
        super().__init__(sequence_length, image_size_pixels, in_channels, out_features)

        # Check that the output shape of the convolutional layers will be at least 1x1
        cnn_spatial_output_size = image_size_pixels - 2 * number_of_conv3d_layers
        if not (cnn_spatial_output_size > 0):
            raise ValueError(
                f"cannot use this many conv3d layers ({number_of_conv3d_layers}) with this input "
                f"spatial size ({image_size_pixels})"
            )

        conv_layers = [
            nn.Conv3d(
                in_channels=in_channels,
                out_channels=conv3d_channels,
                kernel_size=(3, 3, 3),
                padding=(1, 0, 0),
            ),
            nn.LeakyReLU(),
        ]
        if batch_norm:
            # Inserted before activation using position -1
            conv_layers.insert(-1, nn.BatchNorm3d(conv3d_channels))
        for i in range(0, number_of_conv3d_layers - 1):
            conv_layers += [
                nn.Conv3d(
                    in_channels=conv3d_channels,
                    out_channels=conv3d_channels,
                    kernel_size=(3, 3, 3),
                    padding=(1, 0, 0),
                ),
                nn.LeakyReLU(),
            ]
            if batch_norm:
                # Inserted before activation using position -1
                conv_layers.insert(-1, nn.BatchNorm3d(conv3d_channels))

        self.conv_layers = nn.Sequential(*conv_layers)

        # Calculate the size of the output of the 3D convolutional layers
        cnn_output_size = conv3d_channels * cnn_spatial_output_size**2 * sequence_length

        final_block = [
            nn.Linear(in_features=cnn_output_size, out_features=fc_features),
            nn.LeakyReLU(),
            nn.Linear(in_features=fc_features, out_features=out_features),
            nn.LeakyReLU(),
        ]

        if fc_dropout > 0:
            # Insert after the linear layers
            final_block.insert(1, nn.Dropout(fc_dropout))
            final_block.insert(-1, nn.Dropout(fc_dropout))

        self.final_block = nn.Sequential(*final_block)

    def forward(self, x):
        """Run model forward"""
        out = self.conv_layers(x)
        out = out.reshape(x.shape[0], -1)

        # Fully connected layers
        out = self.final_block(out)

        return out


class ResConv3DNet2(AbstractNWPSatelliteEncoder):
    """3D convolutional network based on ResNet architecture.

    The residual blocks are implemented based on the best performing block in [1].

    Sources:
        [1] https://arxiv.org/pdf/1603.05027.pdf
    """

    def __init__(
        self,
        sequence_length: int,
        image_size_pixels: int,
        in_channels: int,
        out_features: int,
        hidden_channels: int = 32,
        n_res_blocks: int = 4,
        res_block_layers: int = 2,
        batch_norm=True,
        dropout_frac=0.0,
    ):
        """Fully connected deep network based on ResNet architecture.

        Args:
            sequence_length: The time sequence length of the data.
            image_size_pixels: The spatial size of the image. Assumed square.
            in_channels: Number of input channels.
            out_features: Number of output features.
            hidden_channels: Number of channels in middle hidden layers.
            n_res_blocks: Number of residual blocks to use.
            res_block_layers: Number of Conv3D layers used in each residual block.
            batch_norm: Whether to include batch normalisation.
            dropout_frac: Probability of an element to be zeroed in the residual pathways.
        """
        super().__init__(sequence_length, image_size_pixels, in_channels, out_features)

        model = [
            nn.Conv3d(
                in_channels=in_channels,
                out_channels=hidden_channels,
                kernel_size=(3, 3, 3),
                padding=(1, 1, 1),
            ),
        ]

        for i in range(n_res_blocks):
            model.extend(
                [
                    ResidualConv3dBlock2(
                        in_channels=hidden_channels,
                        n_layers=res_block_layers,
                        dropout_frac=dropout_frac,
                        batch_norm=batch_norm,
                    ),
                    nn.AvgPool3d((1, 2, 2), stride=(1, 2, 2)),
                ]
            )

        # Calculate the size of the output of the 3D convolutional layers
        final_im_size = image_size_pixels // (2**n_res_blocks)
        cnn_output_size = hidden_channels * sequence_length * final_im_size * final_im_size

        model.extend(
            [
                nn.ELU(),
                nn.Flatten(start_dim=1, end_dim=-1),
                nn.Linear(in_features=cnn_output_size, out_features=out_features),
                nn.ELU(),
            ]
        )

        self.model = nn.Sequential(*model)

    def forward(self, x):
        """Run model forward"""
        return self.model(x)


class EncoderUNET(AbstractNWPSatelliteEncoder):
    """An encoder based on emodifed UNet architecture.

    An encoder for satellite and/or NWP data taking inspiration from the kinds of skip
    connections in UNet. This differs from an actual UNet in that it does not have upsampling
    layers, instead it concats features from different spatial scales, and applies a few extra
    conv3d layers.
    """

    def __init__(
        self,
        sequence_length: int,
        image_size_pixels: int,
        in_channels: int,
        out_features: int,
        n_downscale: int = 3,
        res_block_layers: int = 2,
        conv3d_channels: int = 32,
        dropout_frac: float = 0.1,
    ):
        """An encoder based on emodifed UNet architecture.

        Args:
            sequence_length: The time sequence length of the data.
            image_size_pixels: The spatial size of the image. Assumed square.
            in_channels: Number of input channels.
            out_features: Number of output features.
            n_downscale: Number of conv3d and spatially downscaling layers that are used.
            res_block_layers: Number of residual blocks used after each downscale layer.
            conv3d_channels: Number of channels used in each conv3d layer.
            dropout_frac: Probability of an element to be zeroed in the residual pathways.
        """
        cnn_spatial_output = image_size_pixels // (2**n_downscale)

        if not (cnn_spatial_output > 0):
            raise ValueError(
                f"cannot use this many downscaling layers ({n_downscale}) with this input "
                f"spatial size ({image_size_pixels})"
            )

        super().__init__(sequence_length, image_size_pixels, in_channels, out_features)

        self.first_layer = nn.Sequential(
            nn.Conv3d(
                in_channels=in_channels,
                out_channels=conv3d_channels,
                kernel_size=(1, 1, 1),
                padding=(0, 0, 0),
            ),
            ResidualConv3dBlock(
                in_channels=conv3d_channels,
                n_layers=res_block_layers,
                dropout_frac=dropout_frac,
            ),
        )

        downscale_layers = []
        for _ in range(n_downscale):
            downscale_layers += [
                nn.Sequential(
                    ResidualConv3dBlock(
                        in_channels=conv3d_channels,
                        n_layers=res_block_layers,
                        dropout_frac=dropout_frac,
                    ),
                    nn.ELU(),
                    nn.Conv3d(
                        in_channels=conv3d_channels,
                        out_channels=conv3d_channels,
                        kernel_size=(1, 2, 2),
                        padding=(0, 0, 0),
                        stride=(1, 2, 2),
                    ),
                )
            ]

        self.downscale_layers = nn.ModuleList(downscale_layers)

        self.crop_fn = CenterCrop(cnn_spatial_output)

        cat_channels = conv3d_channels * (1 + n_downscale)
        self.post_cat_conv = nn.Sequential(
            ResidualConv3dBlock(
                in_channels=cat_channels,
                n_layers=res_block_layers,
            ),
            nn.ELU(),
            nn.Conv3d(
                in_channels=cat_channels,
                out_channels=conv3d_channels,
                kernel_size=(1, 1, 1),
            ),
        )

        final_channels = (
            (image_size_pixels // (2**n_downscale)) ** 2 * conv3d_channels * sequence_length
        )
        self.final_layer = nn.Sequential(
            nn.ELU(),
            nn.Linear(
                in_features=final_channels,
                out_features=out_features,
            ),
            nn.ELU(),
        )

    def forward(self, x):
        """Run model forward"""
        out = self.first_layer(x)
        outputs = [self.crop_fn(out)]

        for layer in self.downscale_layers:
            out = layer(out)
            outputs += [self.crop_fn(out)]

        out = torch.cat(outputs, dim=1)
        out = self.post_cat_conv(out)
        out = torch.flatten(out, start_dim=1)
        out = self.final_layer(out)
        return out



"""Basic blocks for image sequence encoders"""
from abc import ABCMeta, abstractmethod

import torch
from torch import nn


class AbstractNWPSatelliteEncoder(nn.Module, metaclass=ABCMeta):
    """Abstract class for NWP/satellite encoder.

    The encoder will take an input of shape (batch_size, sequence_length, channels, height, width)
    and return an output of shape (batch_size, out_features).
    """

    def __init__(
        self,
        sequence_length: int,
        image_size_pixels: int,
        in_channels: int,
        out_features: int,
    ):
        """Abstract class for NWP/satellite encoder.

        Args:
            sequence_length: The time sequence length of the data.
            image_size_pixels: The spatial size of the image. Assumed square.
            in_channels: Number of input channels.
            out_features: Number of output features.
        """
        super().__init__()
        self.out_features = out_features
        self.image_size_pixels = image_size_pixels
        self.sequence_length = sequence_length

    @abstractmethod
    def forward(self):
        """Run model forward"""
        pass


class ResidualConv3dBlock(nn.Module):
    """Fully-connected deep network based on ResNet architecture.

    Internally, this network uses ELU activations throughout the residual blocks.
    """

    def __init__(
        self,
        in_channels,
        n_layers: int = 2,
        dropout_frac: float = 0.0,
    ):
        """Fully-connected deep network based on ResNet architecture.

        Args:
            in_channels: Number of input channels.
            n_layers: Number of layers in residual pathway.
            dropout_frac: Probability of an element to be zeroed.
        """
        super().__init__()

        layers = []
        for i in range(n_layers):
            layers += [
                nn.ELU(),
                nn.Conv3d(
                    in_channels=in_channels,
                    out_channels=in_channels,
                    kernel_size=(3, 3, 3),
                    padding=(1, 1, 1),
                ),
                nn.Dropout3d(p=dropout_frac),
            ]

        self.model = nn.Sequential(*layers)

    def forward(self, x):
        """Run residual connection"""
        return self.model(x) + x


class ResidualConv3dBlock2(nn.Module):
    """Residual block of 'full pre-activation' similar to the block in figure 4(e) of [1].

    This was the best performing residual block tested in the study. This implementation differs
    from that block just by using LeakyReLU activation to avoid dead neurons, and by including
    optional dropout in the residual branch. This is also a 3D fully connected layer residual block
    rather than a 2D convolutional block.

    Sources:
        [1] https://arxiv.org/pdf/1603.05027.pdf
    """

    def __init__(
        self,
        in_channels: int,
        n_layers: int = 2,
        dropout_frac: float = 0.0,
        batch_norm: bool = True,
    ):
        """Residual block of 'full pre-activation' similar to the block in figure 4(e) of [1].

        Sources:
            [1] https://arxiv.org/pdf/1603.05027.pdf

        Args:
            in_channels: Number of input channels.
            n_layers: Number of layers in residual pathway.
            dropout_frac: Probability of an element to be zeroed.
            batch_norm: Whether to use batchnorm
        """
        super().__init__()

        layers = []
        for i in range(n_layers):
            if batch_norm:
                layers.append(nn.BatchNorm3d(in_channels))
            layers.extend(
                [
                    nn.Dropout3d(p=dropout_frac),
                    nn.LeakyReLU(),
                    nn.Conv3d(
                        in_channels=in_channels,
                        out_channels=in_channels,
                        kernel_size=(3, 3, 3),
                        padding=(1, 1, 1),
                    ),
                ]
            )

        self.model = nn.Sequential(*layers)

    def forward(self, x):
        """Run model forward"""
        return self.model(x) + x


class ImageSequenceEncoder(nn.Module):
    """Simple network which independently encodes each image in a sequence into 1D features"""

    def __init__(
        self,
        image_size_pixels: int,
        in_channels: int,
        number_of_conv2d_layers: int = 4,
        conv2d_channels: int = 32,
        fc_features: int = 128,
    ):
        """Simple network which independently encodes each image in a sequence into 1D features.

        For input image with shape [N, C, L, H, W] the output is of shape [N, L, fc_features] where
        N is number of samples in batch, C is the number of input channels, L is the length of the
        sequence, and H and W are the height and width.

        Args:
            image_size_pixels: The spatial size of the image. Assumed square.
            in_channels: Number of input channels.
            number_of_conv2d_layers: Number of convolution 2D layers that are used.
            conv2d_channels: Number of channels used in each conv2d layer.
            fc_features: Number of output nodes for each image in each sequence.
        """
        super().__init__()

        # Check that the output shape of the convolutional layers will be at least 1x1
        cnn_spatial_output_size = image_size_pixels - 2 * number_of_conv2d_layers
        if not (cnn_spatial_output_size >= 1):
            raise ValueError(
                f"cannot use this many conv2d layers ({number_of_conv2d_layers}) with this input "
                f"spatial size ({image_size_pixels})"
            )

        conv_layers = []

        conv_layers += [
            nn.Conv2d(
                in_channels=in_channels,
                out_channels=conv2d_channels,
                kernel_size=3,
                padding=0,
            ),
            nn.ELU(),
        ]
        for i in range(0, number_of_conv2d_layers - 1):
            conv_layers += [
                nn.Conv2d(
                    in_channels=conv2d_channels,
                    out_channels=conv2d_channels,
                    kernel_size=3,
                    padding=0,
                ),
                nn.ELU(),
            ]

        self.conv_layers = nn.Sequential(*conv_layers)

        self.final_block = nn.Sequential(
            nn.Linear(
                in_features=(cnn_spatial_output_size**2) * conv2d_channels,
                out_features=fc_features,
            ),
            nn.ELU(),
        )

    def forward(self, x):
        """Run model forward"""
        batch_size, channel, seq_len, height, width = x.shape

        x = torch.swapaxes(x, 1, 2)
        x = x.reshape(batch_size * seq_len, channel, height, width)

        out = self.conv_layers(x)
        out = out.reshape(batch_size * seq_len, -1)

        out = self.final_block(out)
        out = out.reshape(batch_size, seq_len, -1)

        return out



""" Data module for pytorch lightning """
import glob

from ocf_datapipes.batch import BatchKey, batch_to_tensor, stack_np_examples_into_batch
from ocf_datapipes.training.windnet import windnet_netcdf_datapipe

from pvnet.data.base import BaseDataModule


class WindDataModule(BaseDataModule):
    """Datamodule for training windnet and using windnet pipeline in `ocf_datapipes`."""

    def _get_datapipe(self, start_time, end_time):
        # TODO is this is not right, need to load full windnet pipeline
        data_pipeline = windnet_netcdf_datapipe(
            self.configuration,
            keys=["wind", "nwp", "sensor"],
        )

        data_pipeline = (
            data_pipeline.batch(self.batch_size)
            .map(stack_np_examples_into_batch)
            .map(batch_to_tensor)
        )
        return data_pipeline

    def _get_premade_batches_datapipe(self, subdir, shuffle=False):
        filenames = list(glob.glob(f"{self.batch_dir}/{subdir}/*.nc"))
        data_pipeline = windnet_netcdf_datapipe(
            keys=["wind", "nwp", "sensor"],
            filenames=filenames,
            nwp_channels=self.nwp_channels,
        )

        data_pipeline = (
            data_pipeline.batch(self.batch_size)
            .map(stack_np_examples_into_batch)
            .map(batch_to_tensor)
        )
        if shuffle:
            data_pipeline = (
                data_pipeline.shuffle(buffer_size=100)
                .sharding_filter()
                # Split the batches and reshuffle them to be combined into new batches
                .split_batches(splitting_key=BatchKey.wind)
                .shuffle(buffer_size=self.shuffle_factor * self.batch_size)
            )
        else:
            data_pipeline = (
                data_pipeline.sharding_filter()
                # Split the batches so we can use any batch-size
                .split_batches(splitting_key=BatchKey.wind)
            )

        data_pipeline = (
            data_pipeline.batch(self.batch_size)
            .map(stack_np_examples_into_batch)
            .map(batch_to_tensor)
            .set_length(int(len(filenames) / self.batch_size))
        )

        return data_pipeline



""" Load a model from its checkpoint directory """
import glob
import os

import hydra
import torch
from pyaml_env import parse_config

from pvnet.models.ensemble import Ensemble
from pvnet.models.multimodal.unimodal_teacher import Model as UMTModel


def get_model_from_checkpoints(
    checkpoint_dir_paths: list[str],
    val_best: bool = True,
):
    """Load a model from its checkpoint directory"""
    is_ensemble = len(checkpoint_dir_paths) > 1

    model_configs = []
    models = []
    data_configs = []

    for path in checkpoint_dir_paths:
        # Load the model
        model_config = parse_config(f"{path}/model_config.yaml")

        model = hydra.utils.instantiate(model_config)

        if val_best:
            # Only one epoch (best) saved per model
            files = glob.glob(f"{path}/epoch*.ckpt")
            if len(files) != 1:
                raise ValueError(
                    f"Found {len(files)} checkpoints @ {path}/epoch*.ckpt. Expected one."
                )
            checkpoint = torch.load(files[0], map_location="cpu")
        else:
            checkpoint = torch.load(f"{path}/last.ckpt", map_location="cpu")

        model.load_state_dict(state_dict=checkpoint["state_dict"])

        if isinstance(model, UMTModel):
            model, model_config = model.convert_to_multimodal_model(model_config)

        # Check for data config
        data_config = f"{path}/data_config.yaml"

        if os.path.isfile(data_config):
            data_configs.append(data_config)
        else:
            data_configs.append(None)

        model_configs.append(model_config)
        models.append(model)

    if is_ensemble:
        model_config = {
            "_target_": "pvnet.models.ensemble.Ensemble",
            "model_list": model_configs,
        }
        model = Ensemble(model_list=models)
        data_config = data_configs[0]

    else:
        model_config = model_configs[0]
        model = models[0]
        data_config = data_configs[0]

    return model, model_config, data_config



"""Utils common between Wind and PV datamodules"""
from ocf_datapipes.batch import BatchKey, unstack_np_batch_into_examples
from torch.utils.data import IterDataPipe, functional_datapipe


@functional_datapipe("split_batches")
class BatchSplitter(IterDataPipe):
    """Pipeline step to split batches of data and yield single examples"""

    def __init__(self, source_datapipe: IterDataPipe, splitting_key: BatchKey = BatchKey.gsp):
        """Pipeline step to split batches of data and yield single examples"""
        self.source_datapipe = source_datapipe
        self.splitting_key = splitting_key

    def __iter__(self):
        """Opens the NWP data"""
        for batch in self.source_datapipe:
            for example in unstack_np_batch_into_examples(batch):
                yield example
